from django.conf import settings

RECOMMENDS_STORAGE_REDIS_DATABASE_DEFAULT = {
    'HOST': 'localhost',
    'PORT': 6379,
    'NAME': 0
}

RECOMMENDS_STORAGE_REDIS_DATABASE = getattr(settings, 'RECOMMENDS_STORAGE_REDIS_DATABASE', RECOMMENDS_STORAGE_REDIS_DATABASE_DEFAULT)



import logging
import math
from django.db import transaction
from recommends.storages.base import BaseRecommendationStorage
from recommends.settings import RECOMMENDS_LOGGER_NAME
from .settings import RECOMMENDS_STORAGE_COMMIT_THRESHOLD
from .models import Similarity, Recommendation


logger = logging.getLogger(RECOMMENDS_LOGGER_NAME)


class DjangoOrmStorage(BaseRecommendationStorage):
    def get_similarities_for_object(self, obj, limit=10):
        object_site_id = self.settings.SITE_ID
        return Similarity.objects.similar_to(obj, related_object_site=object_site_id, score__gt=0).order_by('-score')[:limit]

    def get_recommendations_for_user(self, user, limit=10):
        object_site_id = self.settings.SITE_ID
        return Recommendation.objects.filter(user=user.id, object_site=object_site_id).order_by('-score')[:limit]

    def get_votes(self):
        pass

    def store_votes(self, iterable):
        pass

    @transaction.commit_manually
    def store_similarities(self, itemMatch):
        try:
            logger.info('saving similarities')
            count = 0
            for object_id, scores in itemMatch:
                object_target, object_target_site = self.resolve_identifier(object_id)

                for related_object_id, score in scores:
                    if not math.isnan(score) and score > self.threshold_similarities:
                        object_related, object_related_site = self.resolve_identifier(related_object_id)
                        if object_target != object_related:
                            count = count + 1
                            Similarity.objects.set_score_for_objects(
                                object_target=object_target,
                                object_target_site=object_target_site,
                                object_related=object_related,
                                object_related_site=object_related_site,
                                score=score
                            )
                            if count % RECOMMENDS_STORAGE_COMMIT_THRESHOLD == 0:
                                logger.debug('saved %s similarities...' % count)
                                transaction.commit()
        finally:
            logger.info('saved %s similarities...' % count)
            transaction.commit()

    @transaction.commit_manually
    def store_recommendations(self, recommendations):
        try:
            logger.info('saving recommendations')
            count = 0
            for (user, rankings) in recommendations:
                for object_id, score in rankings:
                    if not math.isnan(score) and score > self.threshold_recommendations:
                        count = count + 1
                        object_recommended, site = self.resolve_identifier(object_id)
                        Recommendation.objects.set_score_for_object(
                            user=user,
                            object_recommended=object_recommended,
                            object_site=site,
                            score=score
                        )
                        if count % RECOMMENDS_STORAGE_COMMIT_THRESHOLD == 0:
                            logger.debug('saved %s recommendations...' % count)
                            transaction.commit()
        finally:
            logger.info('saved %s recommendations...' % count)
            transaction.commit()

    def remove_recommendations(self, obj):
        Recommendation.objects.filter_for_object(obj=obj).delete()

    def remove_similarities(self, obj):
        Similarity.objects.filter_for_object(obj=obj).delete()
        Similarity.objects.filter_for_related_object(related_obj=obj).delete()



from django.contrib import admin
from .models import Similarity, Recommendation


class SimilarityAdmin(admin.ModelAdmin):
    list_display = ('object', 'object_site', 'related_object', 'related_object_site', 'score')
    list_filter = ('object_site',)
admin.site.register(Similarity, SimilarityAdmin)


class RecommendationAdmin(admin.ModelAdmin):
    list_display = ('user', 'score', 'object', 'object_site')
    list_filter = ('object_site',)
admin.site.register(Recommendation, RecommendationAdmin)



# Create your views here.
from django.contrib.auth import authenticate, login as _login
from django.http import HttpResponseRedirect


def login(request):
    username = request.POST['username']
    password = request.POST['password']
    user = authenticate(username=username, password=password)
    if user is not None:
        if user.is_active:
            _login(request, user)
            # redirect to a success page.
        else:
            # Return a 'disabled account' error message
            pass
    else:
        pass
        # Return an 'invalid login' error message.
    return HttpResponseRedirect(request.POST.get('next', '/'))



from django.conf import settings


RECOMMENDS_TASK_RUN = getattr(settings, 'RECOMMENDS_TASK_RUN', True)
RECOMMENDS_TASK_CRONTAB = getattr(settings, 'RECOMMENDS_TASK_CRONTAB', {'hour': '*/24'})
RECOMMENDS_CACHE_TEMPLATETAGS_TIMEOUT = getattr(settings, 'RECOMMENDS_CACHE_TEMPLATETAGS_TIMEOUT', 60)
RECOMMENDS_STORAGE_BACKEND = getattr(settings, 'RECOMMENDS_STORAGE_BACKEND', 'recommends.storages.djangoorm.storage.DjangoOrmStorage')
RECOMMENDS_STORAGE_LOGGING_THRESHOLD = getattr(settings, 'RECOMMENDS_STORAGE_LOGGING_THRESHOLD', 1000)
RECOMMENDS_LOGGER_NAME = getattr(settings, 'RECOMMENDS_LOGGER_NAME', 'recommends')



from recommends.converters import IdentifierManager


class BaseRecommendationStorage(object):
    threshold_similarities = 0
    threshold_recommendations = 0

    def __init__(self, settings=None):
        self.identifier_manager = IdentifierManager()
        self.settings = settings

    def get_identifier(self, obj, site_id=None, rating=None, *args, **kwargs):
        """
        Given an object and optional parameters, returns a string identifying the object uniquely.
        """
        if rating is not None:
            site_id = self.get_rating_site(rating).id
        if site_id is None:
            site_id = self.settings.SITE_ID
        return self.identifier_manager.get_identifier(obj, site_id)

    def resolve_identifier(self, identifier):
        """
        This method is the opposite of ``get_identifier``.
        It resolve the object's identifier to an actual model.
        """
        return self.identifier_manager.resolve_identifier(identifier)

    def get_similarities_for_object(self, obj, limit):
        """
        Returns a list of ``Similarity`` objects for ``obj``, ordered by score.
        """
        raise NotImplementedError

    def get_recommendations_for_user(self, user, limit):
        """
        Returns a list of ``Recommendation`` objects for the user, order by score.
        """
        raise NotImplementedError

    def store_similarities(self, itemMatch):
        raise NotImplementedError

    def store_recommendations(self, recommendations):
        """
        Stores all the recommendations.

        ``recommendations`` is an iterable with the following schema:

        ::

            (
                (
                    <user>,
                    (
                        (<object_identifier>, <score>),
                        (<object_identifier>, <score>)
                    ),
                )
            )
        """
        raise NotImplementedError

    def get_votes(self):
        """
        Optional.

        Retrieves the vote matrix saved by ``store_votes``.

        You won't usually need to implement this method, because you want to use fresh data.
        But it might be useful if you want some kind of heavy caching, maybe for testing purposes.
        """
        raise NotImplementedError

    def store_votes(self, iterable):
        """
        Optional.

        Saves the vote matrix.

        You won't usually need to implement this method, because you want to use fresh data.
        But it might be useful if you want to dump the votes on somewhere, maybe for testing purposes.

        ``iterable`` is the vote matrix, expressed as a list of tuples with the following schema:

        ::

            [
                ("<user_id1>", "<object_identifier1>", <score>),
                ("<user_id1>", "<object_identifier2>", <score>),
                ("<user_id2>", "<object_identifier1>", <score>),
                ("<user_id2>", "<object_identifier2>", <score>),
            ]
        """
        raise NotImplementedError

    def remove_recommendation(self, obj):
        """
        Deletes all recommendations for object ``obj``.
        """
        raise NotImplementedError

    def remove_similarity(self, obj):
        """
        Deletes all similarities that have object ``obj`` as source or target.
        """
        raise NotImplementedError



from collections import defaultdict
import math
from recommends.similarities import sim_distance
from recommends.converters import convert_vote_list_to_userprefs, convert_vote_list_to_itemprefs
from .base import BaseAlgorithm


class GhettoAlgorithm(BaseAlgorithm):
    """
    """
    similarity = sim_distance

    def top_matches(self, prefs, p1):
        """
        Returns the best matches for p1 from the prefs dictionary.
        """
        return [(p2, self.similarity(prefs[p1], prefs[p2])) for p2 in prefs if p2 != p1]

    def calculate_similarities(self, vote_list, verbose=0):
        # Invert the preference matrix to be item-centric
        itemPrefs = convert_vote_list_to_itemprefs(vote_list)
        itemMatch = {}
        for item in itemPrefs:
            # Find the most similar items to this one
            itemMatch[item] = self.top_matches(itemPrefs, item)
        iteritems = itemMatch.items()
        return iteritems

    def get_recommended_items(self, vote_list, itemMatch, itemIgnored, user):
        prefs = convert_vote_list_to_userprefs(vote_list)
        itemMatch = dict(itemMatch)

        if user in prefs:
            userRatings = prefs[user]
            scores = defaultdict(int)
            totalSim = defaultdict(int)

            # Loop over items rated by this user
            for (item, rating) in userRatings.iteritems():
                # Loop over items similar to this one
                for (item2, similarity) in itemMatch[item]:
                    # Skip ignored items
                    if user.pk in itemIgnored and item2 in itemIgnored[user.pk]:
                        continue
                    # Ignore if this user has already rated this item
                    if not math.isnan(similarity) and item2 not in userRatings:
                        # Weighted sum of rating times similarity
                        scores[item2] += similarity * rating

                        # Sum of all the similarities
                        totalSim[item2] += similarity

            # Divide each total score by total weighting to get an average
            rankings = ((item, (score / totalSim[item])) for item, score in scores.iteritems() if totalSim[item] != 0)
            return rankings
        return []

    def calculate_recommendations(self, vote_list, itemMatch, itemIgnored):
        """
        ``itemMatch`` is supposed to be the result of ``calculate_similarities()``

        Returns a list of recommendations:

        ::

            [
                (<user1>, [
                    ("<object_identifier1>", <score>),
                    ("<object_identifier2>", <score>),
                ]),
                (<user2>, [
                    ("<object_identifier1>", <score>),
                    ("<object_identifier2>", <score>),
                ]),
            ]

        """
        recommendations = []
        users = set(map(lambda x: x[0], vote_list))
        for user in users:
            rankings = self.get_recommended_items(vote_list, itemMatch, itemIgnored, user)
            recommendations.append((user, rankings))
        return recommendations



import logging
from django.contrib.auth.models import User
from django.contrib.sites.models import Site
from django.conf import settings
from ..converters import model_path
from ..settings import RECOMMENDS_STORAGE_BACKEND, RECOMMENDS_LOGGER_NAME
from ..tasks import remove_suggestions, remove_similarities
from ..utils import import_from_classname
from ..algorithms.ghetto import GhettoAlgorithm


logger = logging.getLogger(RECOMMENDS_LOGGER_NAME)


class RecommendationProviderRegistry(object):
    _vote_providers = {}
    _content_providers = {}
    providers = set()

    def __init__(self):
        StorageClass = import_from_classname(RECOMMENDS_STORAGE_BACKEND)
        self.storage = StorageClass(settings)

    def register(self, vote_model, content_models, Provider):
        provider_instance = Provider()
        self._vote_providers[model_path(vote_model)] = provider_instance
        for model in content_models:
            self._content_providers[model_path(model)] = provider_instance

        self.providers.add(provider_instance)

        for signal in provider_instance.rate_signals:
            if isinstance(signal, str):
                sig_class_name = signal.split('.')[-1]
                sig_instance = import_from_classname(signal)
                listener = getattr(provider_instance, sig_class_name, False)
                if listener:
                    for model in content_models:
                        sig_instance.connect(listener, sender=model)

    def unregister(self, vote_model, content_models, Provider):
        provider_instance = Provider()

        for signal in provider_instance.rate_signals:
            if isinstance(signal, str):
                sig_class_name = signal.split('.')[-1]
                sig_instance = import_from_classname(signal)
                listener = getattr(provider_instance, sig_class_name, False)
                if listener:
                    for model in content_models:
                        sig_instance.disconnect(listener, sender=model)

        new_set = [i for i in self.providers if not isinstance(i, Provider)]
        self.providers = set(new_set)

        for model in content_models:
            del self._content_providers[model_path(model)]
        del self._vote_providers[model_path(vote_model)]

    def get_provider_for_vote(self, model):
        return self._vote_providers[model_path(model)]

    def get_provider_for_content(self, model):
        return self._content_providers[model_path(model)]

    def get_vote_providers(self):
        return self._vote_providers.values()

recommendation_registry = RecommendationProviderRegistry()


class Rating(object):
    def __init__(self, user, rated_object, rating):
        self.user = user
        self.rated_object = rated_object
        self.rating = rating


class RecommendationProvider(object):
    """
    A ``RecommendationProvider`` specifies how to retrieve various informations (items, users, votes)
    necessary for computing recommendation and similarities for a set of objects.

    Subclasses override methods in order to determine what constitutes voted items, a vote,
    its score, and user.
    """
    rate_signals = ['django.db.models.signals.pre_delete']
    algorithm = GhettoAlgorithm()

    def __init__(self):
        if not getattr(self, 'storage', False):
            self.storage = recommendation_registry.storage

    def get_items(self):
        """Return items that have been voted."""
        raise NotImplementedError

    def get_ratings(self, obj):
        """Returns all ratings for given item."""
        raise NotImplementedError

    def get_rating_user(self, rating):
        """Returns the user who performed the rating."""
        raise NotImplementedError

    def get_rating_score(self, rating):
        """Returns the score of the rating."""
        raise NotImplementedError

    def get_rating_item(self, rating):
        """Returns the rated object."""
        raise NotImplementedError

    def get_rating_site(self, rating):
        """Returns the site of the rating. Can be a ``Site`` object or its ID.

        Defaults to ``settings.SITE_ID``."""
        return settings.SITE_ID

    def is_rating_active(self, rating):
        """Returns if the rating is active."""
        return True

    def pre_delete(self, sender, instance, **kwargs):
        """
        This function gets called when a signal in ``self.rate_signals`` is
        fired from one of the rated model.

        Overriding this method is optional. The default method removes the
        suggestions for the deleted objected.

        See :doc:`signals`.
        """
        remove_similarities.delay(rated_model=model_path(sender), object_id=instance.id)
        remove_suggestions.delay(rated_model=model_path(sender), object_id=instance.id)

    def vote_list(self):
        vote_list = self.storage.get_votes()
        if vote_list is None:
            vote_list = []
            for item in self.get_items():
                for rating in self.get_ratings(item):
                    user = self.get_rating_user(rating)
                    score = self.get_rating_score(rating)
                    site = self.get_rating_site(rating)
                    if isinstance(site, Site):
                        site_id = site.id
                    else:
                        site_id = site
                    identifier = self.storage.get_identifier(item, site_id)
                    vote_list.append((user, identifier, score))
            self.storage.store_votes(vote_list)
        return vote_list

    def items_ignored(self):
        """
        Returns user ignored items.
        User can delete items from the list of recommended.

        See recommends.converters.IdentifierManager.get_identifier for help.
        ::

            {<user1>: ["object_identifier1",..., "object_identifierN"], ..}
        """
        return {}

    def precompute(self, vote_list=None):
        """
        This function will be called by the task manager in order
        to compile and store the results.
        """
        if vote_list is None:
            logger.info('fetching votes from the provider...')
            vote_list = self.vote_list()
        logger.info('calculating similarities...')
        self.algorithm.clear_cache()
        itemMatch = self.algorithm.calculate_similarities(vote_list)

        logger.info('saving similarities...')
        self.pre_store_similarities(itemMatch)
        self.storage.store_similarities(itemMatch)

        logger.info('fetching ignored items...')
        itemIgnored = self.items_ignored()

        logger.info('saving recommendations...')
        self.storage.store_recommendations(self.algorithm.calculate_recommendations(
            vote_list, itemMatch, itemIgnored)
        )

    def get_users(self):
        """Returns all users who have voted something."""
        return User.objects.filter(is_active=True)

    def pre_store_similarities(self, itemMatch):
        """
        Optional. This method will get called right before passing the
        similarities to the storage. For example, you can override this method
        to do some stats or visualize the data.
        """
        pass



from django.core.management.base import NoArgsCommand
from recommends.tasks import recommends_precompute


class Command(NoArgsCommand):
    def handle(self, *args, **options):
        recommends_precompute()



from __future__ import unicode_literals
from django.contrib.contenttypes.models import ContentType
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class MockModel(object):
    _object = None

    def __init__(self, **kwargs):
        self.__dict__ = kwargs

    def __str__(self):
        return "Mock for content object %s" % self.object

    def __repr__(self):
        return "<%s>" % self.__str__()

    @property
    def object(self):
        if self._object is None:
            ModelClass = ContentType.objects.get(pk=self.object_ctype).model_class()
            self._object = ModelClass.objects.get(pk=self.object_id)
        return self._object


@python_2_unicode_compatible
class MockSimilarity(MockModel):
    _related_object = None

    @property
    def related_object(self):
        if self._related_object is None:
            ModelClass = ContentType.objects.get(pk=self.related_object_ctype).model_class()
            self._related_object = ModelClass.objects.get(pk=self.related_object_id)
        return self._related_object

    def __str__(self):
        return "Similarity between %s and %s" % (self.object, self.related_object)



#!/usr/bin/env python
import sys
from os import path

from django.conf import settings

PROJECT_DIR = path.dirname(path.realpath(__file__))


settings.configure(
    DATABASES={
        'default': {'ENGINE': 'django.db.backends.sqlite3', 'NAME': ':memory;'}
    },
    INSTALLED_APPS=[
        'django.contrib.auth',
        'django.contrib.sessions',
        'django.contrib.contenttypes',
        'django.contrib.sites',
        'recommends',
        'recommends.storages.djangoorm',
        'recommends.tests',
    ],
    ROOT_URLCONF='recommends.tests.urls',
    TEMPLATE_DIRS=(
        path.join(PROJECT_DIR, 'templates'),
    ),
    BROKER_URL = 'redis://localhost:6379/0',
    CELERY_ALWAYS_EAGER = True,
    ALLOWED_HOSTS = ['*'],
    SITE_ID = 1,
    RECOMMENDS_TEST_REDIS = True,
    RECOMMENDS_TEST_MONGO = True,
    RECOMMENDS_TEST_RECSYS = True
)


def runtests(*test_args):
    import django.test.utils
    runner_class = django.test.utils.get_runner(settings)
    test_runner = runner_class(verbosity=1, interactive=True)
    failures = test_runner.run_tests(['recommends'])
    sys.exit(failures)

if __name__ == '__main__':
    runtests()



import os
from setuptools import setup, find_packages

from recommends import VERSION


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

requirements = [
    'celery',
]

setup(
    name = "django-recommends",
    version = ".".join(map(str, VERSION)),
    description = "A django app that builds item-based suggestions for users",
    long_description = read('README.rst'),
    url = 'https://github.com/python-recsys/django-recommends',
    license = 'MIT',
    author = 'Flavio Curella',
    author_email = 'flavio.curella@gmail.com',
    packages = find_packages(exclude=['tests']),
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Framework :: Django',
        'Topic :: Scientific/Engineering :: Information Analysis',
    ],
    install_requires = requirements,
)



from __future__ import unicode_literals
from django.db import models
from django.contrib.contenttypes import generic
from .managers import RecommendsManager, SimilarityManager, RecommendationManager
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class RecommendsBaseModel(models.Model):
    """(RecommendsBaseModel description)"""
    object_ctype = models.PositiveIntegerField()
    object_id = models.PositiveIntegerField()
    object_site = models.PositiveIntegerField()
    object = generic.GenericForeignKey('object_ctype', 'object_id')

    objects = RecommendsManager()

    class Meta:
        abstract = True
        unique_together = ('object_ctype', 'object_id', 'object_site')

    def __str__(self):
        return "RecommendsBaseModel"


@python_2_unicode_compatible
class Similarity(RecommendsBaseModel):
    """How much an object is similar to another"""

    score = models.FloatField(null=True, blank=True, default=None)

    related_object_ctype = models.PositiveIntegerField()
    related_object_id = models.PositiveIntegerField()
    related_object_site = models.PositiveIntegerField()
    related_object = generic.GenericForeignKey('related_object_ctype', 'related_object_id')

    objects = SimilarityManager()

    class Meta:
        verbose_name_plural = 'similarities'
        unique_together = ('object_ctype', 'object_id', 'object_site', 'related_object_ctype', 'related_object_id', 'related_object_site')
        ordering = ['-score']

    def __str__(self):
        return "Similarity between %s and %s" % (self.object, self.related_object)


@python_2_unicode_compatible
class Recommendation(RecommendsBaseModel):
    """Recommended an object for a particular user"""
    user = models.PositiveIntegerField()
    score = models.FloatField(null=True, blank=True, default=None)

    objects = RecommendationManager()

    class Meta:
        unique_together = ('object_ctype', 'object_id', 'user')
        ordering = ['-score']

    def __str__(self):
        return "Recommendation for user %s" % (self.user)



from recommends.managers import DictStorageManager


class RedisStorageManager(DictStorageManager):
    def similarity_for_objects(self, score, *args, **kwargs):
        spec = super(RedisStorageManager, self).similarity_for_objects(*args, **kwargs)
        spec['score'] = score
        return spec

    def filter_for_object(self, obj):
        ctype_id = self.get_ctype_id_for_obj(obj)
        return {'object_ctype': ctype_id, 'object_id': obj.id}

    def filter_for_related_object(self, related_obj):
        ctype_id = self.get_ctype_id_for_obj(related_obj)
        return {'related_object_ctype': ctype_id, 'related_object_id': related_obj.id}
